from numpy import linalg, zeros, ones, hstack, asarray, vstack, array, mean, std
import itertools
import matplotlib.pyplot as plt
from datetime import datetime
import pandas as pd
import numpy as np
import scipy
import matplotlib.dates as mdates
from sklearn.metrics import mean_squared_error
from math import sqrt
import pvlib
from pvlib import clearsky, atmosphere, solarposition
from pvlib.location import Location
from pvlib.iotools import read_tmy3
from pvlib.irradiance import get_extra_radiation
import warnings
import time
warnings.filterwarnings("ignore")

def classify_weather_day_GM_Tina(df, clearsky_ghi_tag, meas_ghi_tag):
    X_csm = array(df[clearsky_ghi_tag].tolist())
    X_meas = array(df[meas_ghi_tag].tolist())

    # Calculate index of serenity, k
    # a k is calculated for every index

    k = abs(X_meas - X_csm) / X_csm
    k = array(k)
    #print("final vals k: ", k, "min k: ", k.min(), "max k: ", k.max())

    #TODO:
    # 1. CHANGE FREQUENCY TO 1 min/10 min
    # 2. For every Δt (10 minutes/1 hour), divide it into 3 splits
    # 3. For each split, iterate

    # Moving average
    # MA is calculated for a larger interval, i assume
    Nm = 3 # 5
    MA = []
    for i in range(len(k)):

        sumk = 0
        if i < Nm:
            for j in range(i+1):
                sumk += k[j]

        else:
            for iter in range(Nm):
                sumk += k[i - iter]

        MA.append(sumk)

    MA = array(MA) * (1 / Nm)

    #print("CHECK: make sure lengths are equal (k and MA): ", len(k), len(MA))

    # Moving function
    MF = []
    for i in range(len(k)):
        sumMF = 0
        if i < Nm:
            for j in range(i+1):
                sumMF += abs(MA[i] - k[i-iter])

        else:
            for iter in range(Nm):
                # MA does not iter
                sumMF += abs(MA[i] - k[i-iter])

        MF.append(sumMF)

    MF = array(MF)

    # Classification logic
    classification = []
    # input k and MF
    for i in range(len(k)):
        if(MF[i] > 0.05):
            # Variable
            classification.append(1)
        # k[i] = 0.4
        elif (k[i] > 0.7):
            # Cloudy
            classification.append(2)
        elif (k[i] > 0.2 or MF[i] > 0.02):
            # Slightly Cloudy
            classification.append(3)
        else:
            # Clear
            classification.append(4)

    return classification, k, MF

def data_preprocessing(df, xs, Y_tag, I_tag, cs_tag, Y_high_filter, print_info):

    # data processing
    df.dropna(inplace = True)

    df = df[df[xs[0]] > 20]
    # drop where ghi_clearsky is equal to 0 because we will be dividing by that later
    df = df[df[cs_tag] != 0]
    df = df[df[Y_tag] > 0]

    df = df[df[Y_tag] < Y_high_filter]

    # irradiance and temperature sensor verification
    # find all points outside of 3 sigma of Isc / Irradiance
    # replacing Isc with DC Current because no IV trace at inverter level
    # OUTLIER REMOVAL
    old_num_rows = len(df.index)
    I_vs_Irr = array(df[I_tag].tolist()) / array(df[xs[0]].tolist())
    avg = mean(I_vs_Irr, axis = 0)
    sd = std(I_vs_Irr, axis = 0)
    sigmas = 3
    outliers = [(True if ((x < avg - sigmas * sd) or (x > avg + sigmas * sd)) else False) for x in I_vs_Irr]

    df['outlier_bool'] = outliers
    df = df[df['outlier_bool'] == False]
    df.drop(columns = ['outlier_bool'])

    if print_info:
        new_num_rows = len(df.index)
        print("Dropped {} of {} rows with I/Irr filter.".format((old_num_rows - new_num_rows), old_num_rows))

    return df

def add_ghi_to_df(df, start, end, freq, dropped_days, xs, ghi_tag, cs_tag, type_ = None):
    if type_ == 'NIST':
        # multiply GHI by 10^3 because it is in milli
        df[ghi_tag] = df[ghi_tag].apply(lambda x: x * 10**2)

    #TODO: Classify and group days/or/hours (check regression_notes.txt on Desktop)
    if type_ == '8157UCF' or 'PVLifetime' or 'VCAD':
        cocoa = Location(28.387566, -80.755984, tz='US/Eastern', altitude = 10.97)
    elif type_ == 'NIST':
        cocoa = Location(39.994425, -105.265645, tz='US/Mountain', altitude = 1623.974)
    else:
        print('No valid input.')
    times = pd.DatetimeIndex(start=datetime.strptime(start, '%m/%d/%Y %I:%M:%S %p').strftime('%Y-%m-%d'),
                            end=datetime.strptime(end, '%m/%d/%Y %I:%M:%S %p').strftime('%Y-%m-%d'),
                            freq=freq+'in', tz=cocoa.tz)
    cs = cocoa.get_clearsky(times)

    df = df[df[xs[0]] > 20]
    cs = cs[cs['ghi'] > 20]
    cs = cs.iloc[1:]
    cs_ghi = pd.DataFrame()
    cs_ghi[cs_tag] = cs['ghi']

    cs_ghi.index = pd.to_datetime(cs.index, format = '%m/%d/%Y %I:%M:%S %p').strftime('%m/%d/%Y %I:%M:%S %p')

    cs = pd.merge(df, cs_ghi, how='inner', left_index = True, right_index = True)
    df = cs

    if str(type(df.index[0])) != "<class 'str'>":
        df_index = [i.strftime('%m/%d/%Y %I:%M:%S %p') for i in df.index]
        df.index = df_index

    return df



from numpy import linalg, zeros, ones, hstack, asarray, vstack, array, mean, std
import itertools
import matplotlib.pyplot as plt
from datetime import datetime
import pandas as pd
import numpy as np
import matplotlib.dates as mdates
from sklearn.metrics import mean_squared_error
from math import sqrt
import warnings
import copy
import time
warnings.filterwarnings("ignore")

from PVPolyfit import preprocessing as preprocess
from PVPolyfit import utilities
from PVPolyfit import clustering as cluster
from PVPolyfit import kernel


def pvpolyfit(train_df, test_df, Y_tag, xs, I_tag, ghi_tag, cs_tag, highest_num_clusters, highest_degree, kernel_type, Y_high_filter, min_count_per_day, plot_graph = True, graph_type = 'regression', print_info = False):
    pvpoly = PVPolyfit(train_df, test_df, Y_tag, xs, I_tag, ghi_tag, cs_tag, print_info)

    # .prepare has its own filters in place, see docs for information
    pvpoly.prepare(10000, 8)

    rmse_list = []
    std_rmse_list = []
    pvpoly_objects = []
    for i in range(1, highest_num_clusters+1):
        pvpoly_iter = copy.deepcopy(pvpoly)
        try:
            pvpoly_iter.run(num_clusters = i, num_iterations = 1, degrees = list(range(1,highest_degree+1)), kernel_type = kernel_type)
            all_best_dfs, ultimate_days, avg_rmse, std_rmse = pvpoly_iter.evaluate(print_info = print_info)
            rmse_list.append(avg_rmse)
            std_rmse_list.append(std_rmse)
            pvpoly_objects.append(pvpoly_iter)

        except:
            break

    if len(rmse_list) == 0:
        raise Exception("No Output was produced. Check to make sure your inputs follow the documentation.")

    min_idx = np.argmin(rmse_list)

    if print_info:
        print("{} cluster(s) were used.".format(range(1,highest_num_clusters)[min_idx]))

    days_rmses, model_output = pvpoly_objects[min_idx].plot(graph_type = graph_type, print_info = print_info, plot_graph = plot_graph)

    return model_output, days_rmses


class PVPolyfit:
    '''
    .______   ____    ____ .______     ______    __      ____    ____  _______  __  .___________.
    |   _  \  \   \  /   / |   _  \   /  __  \  |  |     \   \  /   / |   ____||  | |           |
    |  |_)  |  \   \/   /  |  |_)  | |  |  |  | |  |      \   \/   /  |  |__   |  | `---|  |----`
    |   ___/    \      /   |   ___/  |  |  |  | |  |       \_    _/   |   __|  |  |     |  |     
    |  |         \    /    |  |      |  `--'  | |  `----.    |  |     |  |     |  |     |  |     
    | _|          \__/     | _|       \______/  |_______|    |__|     |__|     |__|     |__|     

    An object, PVPolyfit, created for the creation of an accurate regression of Output depending on the two covariates

    PARAMETERS
    ----------

        train_df: df
            holds training data with columns and index specified below
        test_df: df
            holds testing data with columns and index specified below
        Y_tag: str
            column name of output tag
        xs: list of str
            list of column names for two covariates
        ghi_tag: str
            column name of GHI input
        cs_tag: str
            column name of clearsky GHI generated by pvlib simulation (link below)

    USER MUST INPUT DF's WITH FOLLOWING COLUMNS:

    |       Description      |  Original Use Case  |   Model Purpose          |
    |------------------------|---------------------|--------------------------|
    | Output,       Y_tag:   | DC Power            | Target for regression    |
    | xs:              x1:   | POA Irradiance      | Covariate for regression |
    |                  x2:   | Ambient Temperature | Covariate for regression |
    | Measured GHI, ghi_tag  | GHI (irradiance)    | Day classification       |
    | PVLib Clearsky, cs_tag | Simulated GHI       | Day classification       |

    PVLib has a good tutorial to generate clearsky data: https://pvlib-python.readthedocs.io/en/stable/generated/pvlib.location.Location.get_clearsky.html
    '''

    def __init__(self, train_df, test_df, Y_tag, xs, I_tag, ghi_tag, cs_tag, print_info):
        self.train_df = train_df
        self.test_df = test_df
        self.Y_tag = Y_tag
        self.xs = xs
        self.I_tag = I_tag
        self.ghi_tag = ghi_tag
        self.cs_tag = cs_tag
        self.print_info = print_info


        self.num_clusters = 0
        self.num_iterations = 0
        self.degrees = []

        self.cut_results = []
        self.test_cut_results = []
        self.ordered_pair_list = []
        self.test_ordered_pair_list = []
        self.combined_test_cut_results = []
        self.test_km_labels = []
        self.avg_rmse = 0


        # all_best_dfs[Degree][Day][Columns: 'Y', 'mins', 'maxs']
        self.all_best_dfs = []

        # model_day_couts[Degree][Best model's [Train Counts, Test Counts]]
        self.model_day_counts = []

        # ultimate_days[Day][i, ind]
        self.ultimate_days = []

    def prepare(self, Y_high_filter, min_count_per_day):
        ''' Preprocess and classify days in DataFrame '''
        self.train_df = preprocess.data_preprocessing(self.train_df, self.xs, self.Y_tag, self.I_tag, self.cs_tag, Y_high_filter, self.print_info)
        classification, k, MF = preprocess.classify_weather_day_GM_Tina(self.train_df, self.cs_tag, self.ghi_tag)

        self.train_df['day_type'] = classification

        # cuts train_df into daily DF's
        # also returns a filtered train_df which cuts out a day if its length is too small
        index_list, day_hour_list, self.cut_results, self.train_df = utilities.find_and_break_days_or_hours(self.train_df, True, min_count_per_day = min_count_per_day, frequency = 'days', print_info = self.print_info)

        middles_dates, hours_kpi = utilities.get_weighted_middle_of_day_and_calculate_float_since_noon(self.cut_results, self.ghi_tag, self.cs_tag)
        ### For each day, compile frequencies
        ### For each day, output the # times each class is triggered
        self.ordered_pair_list, freq_df = cluster.create_conglomerated_vectors_for_clustering_algorithm(self.cut_results, hours_kpi, day_hour_list, self.Y_tag)

        self.test_df = preprocess.data_preprocessing(self.test_df, self.xs, self.Y_tag, self.I_tag, self.cs_tag, Y_high_filter, self.print_info)
        test_classification, test_k, test_MF = preprocess.classify_weather_day_GM_Tina(self.test_df, self.cs_tag, self.ghi_tag)
        self.test_df['day_type'] = test_classification
        test_index_list, test_day_hour_list, self.test_cut_results, self.test_df = utilities.find_and_break_days_or_hours(self.test_df, True, min_count_per_day = min_count_per_day, frequency = 'days', print_info = self.print_info)
        test_middles_dates, test_hours_kpi = utilities.get_weighted_middle_of_day_and_calculate_float_since_noon(self.test_cut_results, self.ghi_tag, self.cs_tag)
        self.test_ordered_pair_list, test_freq_df = cluster.create_conglomerated_vectors_for_clustering_algorithm(self.test_cut_results, test_hours_kpi, test_day_hour_list, self.Y_tag)

    def run(self, num_clusters = 6, num_iterations = 1, degrees = list(range(1,10)), kernel_type = 'polynomial'):
        '''
        Iterates through Degrees
        For each Degree, iterates n times
        Returns best model for each input day

        Parameters:
            num_clusters: int, default 6
                number of clusters used in clustering algorithm, synonymous with number of 'types of days'

            num_iterations: int, default 1
                number of times algorithm loops, indicates volatility of algorithm (usually very small, so default = 1)

            degrees: list of ints
                range of degrees that polynomial kernel iterates through

            kernel_type: str
                type of regression kernel to be used
                OPTIONS: polynomial - a(AB)+
        '''

        self.num_clusters = num_clusters
        self.num_iterations = num_iterations
        self.degrees = degrees
        self.kernel_type = kernel_type

        self.all_best_dfs = []
        self.model_day_counts = []
        for degree in self.degrees:
            P_se_list = []
            combined_P_list = []
            combined_day_counts = []

            # 1. Run the code an n number of times
            for i in range(self.num_iterations):

                # clusters and adds 'model_num' column to cut_results & test_cut_results
                train_kmeans_dfs, test_kmeans_dfs, self.test_km_labels, self.cut_results, self.test_cut_results, train_model_day_count, test_model_day_count = cluster.cluster_ordered_pairs_and_return_df_of_days_in_cluster(self.cut_results, self.test_cut_results, self.ordered_pair_list, self.test_ordered_pair_list, kmeans_num_clusters = self.num_clusters, print_info = self.print_info)
                saved_models = cluster.save_model_for_each_cluster(train_kmeans_dfs, degree, self.Y_tag, self.xs, self.kernel_type)
                self.kmeans_Y_lists = kernel.process_test_data_through_models(test_kmeans_dfs, saved_models, self.test_km_labels, self.xs)

                # 2. For each iteration, save the modelled P and colors (based on model used)
                combined_P_list.append(self.kmeans_Y_lists)
                self.combined_test_cut_results.append(self.test_cut_results)
                combined_day_counts.append([train_model_day_count, test_model_day_count])
                P_se_km = kernel.EvaluateModel(array(self.test_df[self.Y_tag].tolist()),array(self.kmeans_Y_lists)).rmse()
                P_se_list.append(P_se_km)

            # 3. Gather the minimum and maximum for each index, save in two lists
            mins = []
            maxs = []
            for i in range(len(self.test_df.index)):
                min = 9999
                max = -9999
                for j in range(len(combined_P_list)):
                    if (combined_P_list[j][i] < min):
                        min = combined_P_list[j][i]
                    if (combined_P_list[j][i] > max):
                        max = combined_P_list[j][i]
                mins.append(min)
                maxs.append(max)

            best_index = np.argmin(P_se_list)
            best_model = combined_P_list[best_index]
            best_df = pd.DataFrame()

            best_df['Y'] = best_model
            best_df['mins'] = mins
            best_df['maxs'] = maxs

            best_df.index = self.test_df.index
            _,_,dfg,_ = utilities.find_and_break_days_or_hours(best_df, False, min_count_per_day = 0, frequency = 'days', print_info = self.print_info)
            self.all_best_dfs.append(dfg)
            self.model_day_counts.append(combined_day_counts[best_index])

    def evaluate(self, print_info = True):
        '''
            Determine rmse for each day for each degree
            and return index of best model for each day
        '''
        # iterate by day
        all_rmse = []
        self.ultimate_days = []
        for i in range(len(self.all_best_dfs[0])):
            min = 9999
            ind = 0
            # iterate by degree
            for j in range(len(self.all_best_dfs)):
                iterating_rmse = kernel.EvaluateModel(array(self.test_cut_results[i][self.Y_tag].tolist()),array(self.all_best_dfs[j][i]['Y'].tolist())).rmse()
                if abs(iterating_rmse) < abs(min):
                    min = iterating_rmse
                    ind = j
            if print_info:
                print("{} index: {}, degrees len: {}".format(len(self.all_best_dfs), ind, len(self.degrees)))
                print("Day {} chooses degree {} with {}".format(i, self.degrees[ind], min))
            all_rmse.append(min)
            self.ultimate_days.append([i, ind])

        self.avg_rmse = np.array(all_rmse).mean()
        return self.all_best_dfs, self.ultimate_days, self.avg_rmse, np.array(all_rmse).std()

    def plot(self, graph_type = 'regression', print_info = True, plot_graph = False):
        iter_rmses = []
        if graph_type == 'regression':
            colors = ['red', 'blue', 'green', 'orange', 'purple', 'brown', 'gold', 'pink', 'gray', 'cyan', 'darkgreen', 'burleywood', 'lawngreen', 'cornflowerblue']*5
            model_outputs = []
            for i in range(len(self.all_best_dfs[0])):
                model_number = self.test_km_labels[i]
                color = colors[model_number]

                ind = self.ultimate_days[i][1]

                Y_output_daily = self.all_best_dfs[ind][i]['Y'].tolist()
                model_outputs.append(Y_output_daily)
                day_index = self.all_best_dfs[ind][i].index.tolist()
                day_maxes = self.all_best_dfs[ind][i]['maxs'].tolist()
                day_mins = self.all_best_dfs[ind][i]['mins'].tolist()
                day_meas = array(self.test_cut_results[i][self.Y_tag].tolist())
                dt_index = pd.to_datetime(day_index)

                if plot_graph:
                    plt.plot(dt_index, day_meas, 'k')
                    plt.plot(dt_index, Y_output_daily, color)
                    plt.fill_between(dt_index, day_maxes, day_mins, facecolor = color)
                    plt.xlabel("time")
                    plt.ylabel("Watts")
                    plt.xticks(rotation=60)
                    plt.title('Modelled Multiple Day Types (by color)')


                uncer = (array(Y_output_daily)-day_meas)#/(day_meas))
                calc_rmse = sqrt(mean_squared_error(day_meas, array(Y_output_daily)))
                iter_rmses.append(calc_rmse)

                if print_info:
                    print("[{}]:".format(datetime.strptime(day_index[0], '%m/%d/%Y %H:%M:%S %p').strftime('%Y-%m-%d')))
                    print("\trmse: {:.4f}, error: {:.4f}".format(calc_rmse, uncer.mean()))
            if plot_graph:
                plt.show()
        return iter_rmses, model_outputs

    def plot_classification_map(self):
        return

    def model_information(self):
        return



from PVPolyfit import kernel

from numpy import linalg, zeros, ones, hstack, asarray, vstack, array, mean, std
import pandas as pd

def cluster_ordered_pairs_and_return_df_of_days_in_cluster(cut_results, test_cut_results, ordered_pair_list, test_ordered_pair_list, kmeans_num_clusters = 4, print_info = False):
    ''' KNN Clustering Algorithm - to find the same types of days'''
    from sklearn.cluster import KMeans

    kmeans = KMeans(n_clusters = kmeans_num_clusters)
    kmeans.fit(array(ordered_pair_list + test_ordered_pair_list))

    centroids = kmeans.cluster_centers_
    km_labels = kmeans.labels_
    #print('labels, centroids', km_labels, centroids)

    train_km_labels = km_labels[:len(ordered_pair_list)]
    test_km_labels = km_labels[len(ordered_pair_list):]

    # For cut_results, make new column with each cluster the day belongs to
    for i in range(len(cut_results)):
        cut_results[i]['model_num'] = train_km_labels[i]
    for i in range(len(test_cut_results)):
        test_cut_results[i]['model_num'] = test_km_labels[i]

    train_array_km_labels = array(train_km_labels)
    test_array_km_labels = array(test_km_labels)

    #for i in range(kmeans_num_clusters):
    #    print("[train]: cluster num ", i, " has ", len(train_array_km_labels[train_array_km_labels == i]))

    #for i in range(kmeans_num_clusters):
    #    print("[test ]: cluster num ", i, " has ", len(test_array_km_labels[test_array_km_labels == i]))

    #print("LEN(train_cut_results): ", len(cut_results))
    train_kmeans_dfs = [pd.DataFrame() for i in range(kmeans_num_clusters)]
    train_model_day_count = [0] * kmeans_num_clusters
    for i in range(len(cut_results)):
        train_model_day_count[train_km_labels[i]] += 1
        train_kmeans_dfs[train_km_labels[i]] = pd.concat([train_kmeans_dfs[train_km_labels[i]], cut_results[i]])

    if print_info:
        print("[TRAIN]: NUM DAYS PER MODEL", train_model_day_count)

    test_kmeans_dfs = [pd.DataFrame() for i in range(kmeans_num_clusters)]
    test_model_day_count = [0] * kmeans_num_clusters
    for i in range(len(test_cut_results)):
        test_model_day_count[test_km_labels[i]] += 1
        test_kmeans_dfs[test_km_labels[i]] = pd.concat([test_kmeans_dfs[test_km_labels[i]], test_cut_results[i]])

    if print_info:
        print("[TEST]: NUM DAYS PER MODEL", test_model_day_count)
    #print("LENGTH list of dfs: ", len(test_kmeans_dfs), len(test_km_labels))

    return train_kmeans_dfs, test_kmeans_dfs, test_km_labels, cut_results, test_cut_results, train_model_day_count, test_model_day_count

def save_model_for_each_cluster(kmeans_dfs, degree, Y_tag, xs, kernel_type):
    # each df corresponds to each model

    # save a model for each cluster of days
    saved_models = []
    #print"start of model")
    for i in range(len(kmeans_dfs)):
        #print"len(kmeans_dfs) for model: ", i, " is ", len(kmeans_dfs[i]))
        if len(kmeans_dfs[i]) == 0:
            # if model has no data
            saved_models.append(0)
            continue

        Y = array(kmeans_dfs[i][Y_tag].tolist())

        tup_list = []
        for j in range(len(xs)):
            tup_list.append(array(kmeans_dfs[i][xs[j]].tolist()))

        tup = tuple(tup_list)

        model = kernel.Model(tup, Y, degree, kernel_type)
        
        model.build()

        saved_models.append(model)

    return saved_models

def create_conglomerated_vectors_for_clustering_algorithm(cut_results, hours_kpi, day_hour_list, Y_tag):
    freq_list = []
    # ordered pairs used for kmeans
    ordered_pair_list = []
    big_df = pd.DataFrame(columns = ['variable', 'cloudy', 'slightly cloudy', 'clear', 'hours'])
    dict_list = []
    for l in range(len(cut_results)):
        df = cut_results[l]

        # possible other kpi
        max_v = mean(df[Y_tag].tolist())

        onez, twos, threes, fours = len(df[df['day_type'] == 1]), len(df[df['day_type'] == 2]), len(df[df['day_type'] == 3]), len(df[df['day_type'] == 4])
        if (onez + twos + threes + fours != len(cut_results[l])):
            raise Exception("Sum of classes does not equal length of df")

        # consider adding index here, or the time of day
        freq_df = pd.DataFrame({'variable':onez, 'cloudy':twos, 'slightly cloudy':threes, 'clear':fours}, index = day_hour_list)
        freq_list.append(freq_df)

        ordered_pair = [hours_kpi[l], array(df[Y_tag].tolist()).sum()]
        ordered_pair_list.append(ordered_pair)

        dict = {'variable':onez, 'cloudy':twos, 'slightly cloudy':threes, 'clear':fours, 'hours':hours_kpi[l], 'mean_power':max_v, 'energy':array(df[Y_tag].tolist()).sum()}
        dict_list.append(dict)

    #print("length of dict list and day hour list: ", len(dict_list), len(day_hour_list))
    day_list = list(set(day_hour_list))

    big_df = pd.DataFrame(dict_list, index = day_list)

    #sns.pairplot(big_df)
    #plt.show()
    #print(ordered_pair_list)

    return ordered_pair_list, big_df



from numpy import linalg, zeros, ones, hstack, asarray, vstack, array, mean, std
import itertools
import matplotlib.pyplot as plt
from datetime import datetime
import pandas as pd
import numpy as np
import scipy
import matplotlib.dates as mdates
from sklearn.metrics import mean_squared_error
from math import sqrt
import warnings
warnings.filterwarnings("ignore")



from PVPolyfit import utilities

class Model:

    def __init__(self, inputs, Y, degree, kernel_type):

        self.inputs = inputs
        self.Y = Y
        self.degree = degree

        # possibilities: 'polynomial', 'polynomial with log(POA)', 'diode inspired'
        self.kernel_type = kernel_type

        self.a_hat = []
        self.powers = []

    def build(self):
        """
        Least-squares implementation on multiple covariates
        """

        if self.kernel_type == 0:
            # polynomial

            xs = vstack(self.inputs).T
            num_inputs, len_input = xs.shape[1], xs.shape[0]
            # add column of rows in first index of matrix
            xs = hstack((ones((len_input, 1), dtype=float), xs))

            # construct identity matrix
            iden_matrix = []
            for i in range(num_inputs+1):
                # create array of zeros
                row = zeros(num_inputs+1, dtype=int)
                # add 1 to diagonal index
                row[i] = 1
                iden_matrix.append(row)

            # gather list
            combinations = itertools.combinations_with_replacement(iden_matrix, self.degree)
            # list of polynomial powers
            poly_powers = []
            for i in combinations:
                sum_arr = np.zeros(num_inputs+1, dtype=int)
                for j in i:
                    sum_arr += array(j)
                poly_powers.append(sum_arr)

            # Raise data to specified degree pattern and stack
            A = []
            for power in poly_powers:
                product = (xs**power).prod(1)
                A.append(product.reshape(product.shape + (1,)))
            A = hstack(array(A))

            # get solution with smallest error via least-squares
            # returns coefficients of polynomial
            a_hat = linalg.lstsq(A, self.Y, rcond=-1)[0]

            # check if valid lengths
            if len(a_hat) == 0 or len(poly_powers) == 0:
                raise Exception("PVPolyfit algorithm returned list of length zero for either coeff. or powers")

            # save resolved coefficients 
            self.a_hat = a_hat
            # save polynomial powers
            self.powers = poly_powers

        if self.kernel_type == 1:
            # polynomial with included log(POA) parameter
            # Requires POA be first input in xs

            xs = vstack(self.inputs).T
            num_inputs, len_input = xs.shape[1], xs.shape[0]
            # add column of rows in first index of matrix
            xs = hstack((ones((len_input, 1), dtype=float), xs, vstack(np.log(self.inputs[0]))))

            # construct identity matrix
            iden_matrix = []
            for i in range(num_inputs+1+1):
                # create array of zeros
                row = zeros(num_inputs+1+1, dtype=int)
                # add 1 to diagonal index
                row[i] = 1
                iden_matrix.append(row)

            # gather list
            combinations = itertools.combinations_with_replacement(iden_matrix, self.degree)
            # list of polynomial powers
            poly_powers = []
            for i in combinations:
                sum_arr = np.zeros(num_inputs+1+1, dtype=int)
                for j in i:
                    sum_arr += array(j)
                poly_powers.append(sum_arr)

            # Raise data to specified degree pattern and stack
            A = []
            for power in poly_powers:
                product = (xs**power).prod(1)
                A.append(product.reshape(product.shape + (1,)))
            A = hstack(array(A))

            # get solution with smallest error via least-squares
            # returns coefficients of polynomial
            a_hat = linalg.lstsq(A, self.Y, rcond=-1)[0]

            # check if valid lengths
            if len(a_hat) == 0 or len(poly_powers) == 0:
                raise Exception("PVPolyfit algorithm returned list of length zero for either coeff. or powers")

            # save resolved coefficients 
            self.a_hat = a_hat
            # save polynomial powers
            self.powers = poly_powers


        if self.kernel_type == 2:
            # Diode Inspired
            # Requires that xs inputs be [POA, Temp], in that order

            xs = vstack(self.inputs).T
            num_inputs, len_input = xs.shape[1], xs.shape[0]
            # add column of rows in first index of matrix
            xs = hstack((ones((len_input, 1), dtype=float), xs, np.log(xs)))

            # construct identity matrix
            iden_matrix = []
            for i in range(num_inputs+1+num_inputs):
                # create array of zeros
                row = zeros(num_inputs+1+num_inputs, dtype=int)
                # add 1 to diagonal index
                row[i] = 1
                iden_matrix.append(row)

            A = xs
            # get solution with smallest error via least-squares
            # returns coefficients of polynomial
            self.a_hat = linalg.lstsq(A, self.Y, rcond=-1)[0]
            self.powers = []

    def output(self, temps):
        ''' Evaluate output with input parameters
            and polynomial information 
            
            temps: temporary inputs
            
        '''

        if self.kernel_type == 0:
            #polynomial
            fit = 0
            for b, z in zip(self.a_hat, self.powers):
                iter = b
                for k in range(1, len(z)):
                    iter *= temps[k-1]**z[k]
                fit += iter

        if self.kernel_type == 1:
            #polynomial with included log(POA) parameter
            # requires POA be first input in xs
            
            temps.append(np.log(temps[0]))
            
            fit = 0
            for b, z in zip(self.a_hat, self.powers):
                iter = b
                for k in range(1, len(z)):
                    iter *= temps[k-1]**z[k]
                fit += iter            
    
        if self.kernel_type == 2:
            # diode inspired
            # requires [POA, Temp] as inputs, in that order

            x1_i, x2_i = temps
            b1, b2, b3, b4, b5 = self.a_hat
            fit = b1 + b2*x1_i + b3*x2_i + b4*np.log(x1_i) + b5*np.log(x2_i)            


        return fit

    def info(self):
        return self.a_hat, self.powers

class EvaluateModel:

    def __init__(self, measured, modelled):
        from sklearn.metrics import mean_squared_error
        from math import sqrt

        self.measured = measured
        self.modelled = modelled

    def r_squared(self):
        ''' Calculate model's r-squared value '''
        y_mean_line = [mean(self.measured) for y in self.measured]
        rmse_model = sqrt(mean_squared_error(self.measured, self.modelled))
        rmse_ymean = sqrt(mean_squared_error(self.measured, y_mean_line))
        return 1 - (rmse_model/rmse_ymean)

    def rmse(self):
        ''' Calculate model's Root Mean Square Error (RMSE) '''
        return sqrt(mean_squared_error(self.measured, self.modelled))


def process_test_data_through_models(test_kmeans_dfs, kmeans_saved_models, test_km_labels, xs):
    # When inputted, test_kmeans_dfs is ordered by the number of kmeans clusters
    # This will then be transitioned to be ordered by day
    # Then, it will be pushed through the models
    
    new_dfs = []
    for i in range(len(test_kmeans_dfs)):
        # Check for error case
        if kmeans_saved_models[i] == 0 and len(test_kmeans_dfs[i] != 0):
            raise Exception("Input Error: PVPolyfit requires either less clusters or more training data.")

        if len(test_kmeans_dfs[i]) == 0:
            continue

        # need to parse days from each df
        _, _, dfs, _ = utilities.find_and_break_days_or_hours(test_kmeans_dfs[i], False, min_count_per_day = 0, frequency = 'days')
        new_dfs.append(dfs)

    # flatten list of lists
    test_kmeans_dfs = [item for sublist in new_dfs for item in sublist]

    # sort the dfs by datetime index
    for i in range(len(test_kmeans_dfs)):
        for j in range(len(test_kmeans_dfs)):
            if (datetime.strptime(test_kmeans_dfs[i].index[0], '%m/%d/%Y %H:%M:%S %p') < datetime.strptime(test_kmeans_dfs[j].index[0], '%m/%d/%Y %H:%M:%S %p')):
                temp = test_kmeans_dfs[i]
                test_kmeans_dfs[i] = test_kmeans_dfs[j]
                test_kmeans_dfs[j] = temp

    # iterate through dfs and run models
    kmeans_Y_lists = []

    for i in range(len(test_kmeans_dfs)):
        # if model does not have any days
        if len(test_kmeans_dfs[i]) == 0:
            raise Exception("DataFrame of zero length has been detected")
        
        temps = []
        for j in range(len(xs)):
            temps.append(array(test_kmeans_dfs[i][xs[j]].tolist()))

        model_index = test_km_labels[i]
        Y_list = []
        for j in range(len(temps[0])):
            inputs_iter = [item[j] for item in temps]
            Y_val = (kmeans_saved_models[model_index]).output(inputs_iter)
            Y_list.append(Y_val)

        kmeans_Y_lists.append(Y_list)

    flattened_kmeans_Y_lists = [item for sublist in kmeans_Y_lists for item in sublist]

    return flattened_kmeans_Y_lists


from numpy import linalg, zeros, ones, hstack, asarray, vstack, array, mean, std
import itertools
import matplotlib.pyplot as plt
from datetime import datetime
import pandas as pd
import numpy as np
import scipy
import matplotlib.dates as mdates
from sklearn.metrics import mean_squared_error
from math import sqrt
import warnings
import time
warnings.filterwarnings("ignore")

def find_and_break_days_or_hours(df, filter_bool, min_count_per_day = 8, frequency = 'days', print_info = False):
    index_list = []
    day_hour_list = []
    prev = 0

    for index, j in enumerate(df.index):
        if frequency == 'days':
            curr = int(datetime.strptime(j, '%m/%d/%Y %H:%M:%S %p').strftime('%d'))
            frq = datetime.strptime(j, '%m/%d/%Y %H:%M:%S %p').strftime('%m/%d/%Y')
        elif frequency == 'hours':
            curr = int(datetime.strptime(j, '%m/%d/%Y %H:%M:%S %p').strftime('%H'))
            frq = datetime.strptime(j, '%m/%d/%Y %H:%M:%S %p').strftime('%m/%d/%Y %H')
        if curr != prev:
            index_list.append(index)
            day_hour_list.append(frq)
        prev = curr
        last_index = index

    cut_results = []
    # Break df into days
    for k in range(len(index_list)):
        if k == (len(index_list)-1):
            # append last day
            cut_results.append(df[index_list[k]:-1])
        else:
            cut_results.append(df[index_list[k]:index_list[k+1]])

    cut_results[-1] = pd.concat([cut_results[-1], df.iloc[[-1]]])


    if filter_bool:
        # NUMBER OF POINTS PER DAY MUST BE AT CERTAIN THRESHOLD
        checked_cut_results = []
        checked_index_list = []
        checked_day_hour_list = []
        dropped_days = []
        for i in range(len(cut_results)):
            #TODO: also check standard deviation for flatlining
            if len(cut_results[i]) > min_count_per_day:
                checked_cut_results.append(cut_results[i])
                checked_index_list.append(index_list[i])
                checked_day_hour_list.append(day_hour_list[i])
            else:
                dropped_days.append(cut_results[i].index[0])

        cut_results = checked_cut_results
        index_list = checked_index_list
        day_hour_list = checked_day_hour_list

        if len(dropped_days) > 0:
            df = pd.concat(cut_results)
    
        else:
            if print_info:
                print("No need to alter df because no dropped days detected")

    return index_list, day_hour_list, cut_results, df

def get_weighted_middle_of_day_and_calculate_float_since_noon(cut_results, ghi_tag, cs_tag):
    middles = []
    middles_dates = []
    cur_diff_integral = 999999
    for i in range(len(cut_results)):
        day_diff = array(cut_results[i][cs_tag].tolist()) - array(cut_results[i][ghi_tag].tolist())
        day_start = int(datetime.strptime(cut_results[i].index[0], '%m/%d/%Y %H:%M:%S %p').strftime('%H'))
        for l in range(1,len(day_diff)-1):
            left_data = day_diff[:l]
            right_data = day_diff[l:]
            left_integral = scipy.integrate.simps(left_data)
            right_integral = scipy.integrate.simps(right_data)

            if (abs(left_integral - right_integral) < cur_diff_integral):
                archived_index = l
                archived_index_date = cut_results[i].index[l]
                cur_diff_integral = abs(left_integral - right_integral)

        middles.append(archived_index)
        middles_dates.append(archived_index_date)
        archived_index = 0
        archived_index_date = 0
        cur_diff_integral = 999999

    # For each middle, calculate number of hours, as float, since noon
    hours_kpi = []
    for i in middles_dates:
        noon = datetime.strptime((str(datetime.strptime(i, '%m/%d/%Y %H:%M:%S %p').strftime('%m/%d/%Y')) + ' 12:00:00 PM'), '%m/%d/%Y %H:%M:%S %p')
        cur = datetime.strptime(i, '%m/%d/%Y %I:%M:%S %p')
        td = cur - noon
        float_hours = td.total_seconds() / 3600

        hours_kpi.append(float_hours)

    return middles_dates, hours_kpi





from distutils.core import setup
setup(
  name = 'PVPolyfit',
  packages = ['PVPolyfit'],
  version = '0.3',
  license='MIT',
  description = 'A high-resolution multiple linear regression algorithm used to analyze PV output with a few inputs',
  author = 'Michael Hopwood',
  author_email = 'mwhopwood@gmail.com',
  url = 'https://github.com/MichaelHopwood/PVPolyfit',
  download_url = 'https://github.com/MichaelHopwood/PVPolyfit/archive/v_03.tar.gz',
  keywords = ['Multiple linear regression', 'Multiple linear', 'PV', 'Photovoltaic', 'solar'],
  install_requires=[
          'numpy',
          'matplotlib',
          'datetime',
          'pandas',
          'sklearn',
          'pvlib',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
